"""
Get Generic Type Mixin
======================

"""

from typing import (
    TypeVar,
    Type,
    get_args,
    _GenericAlias
)
from types import GenericAlias #pylint: disable=no-name-in-module

O = TypeVar("O")
OC = TypeVar("OC", bound="GetGenericTypeMixin")

class GetGenericTypeMixin: #pylint: disable=too-few-public-methods
    """
    Allows fetching of class provided generic type
    """

    def __init__(self, *args, **kwargs) -> None:

        super().__init__(*args, **kwargs)

    def _get_generic_type(self: OC) -> Type[O]:
        """Sets the generic type attribute"""

        orig_value = getattr(self, "__orig_bases__", None)
        if orig_value is None:
            orig_value = (getattr(self, "__orig_class__"),)

        # Find generic with mixins
        generic_type = self._search_for_generic_type(orig_value)

        if generic_type is None:
            raise TypeError(f"No generic type found in bases {orig_value}")

        return generic_type

    def _search_for_generic_type(self, orig_value: object):

        generic_type = None

        for base in orig_value:

            print(type(base))

            if isinstance(base, (GenericAlias, _GenericAlias)):
                generic_type = get_args(base)[0]
                break

            if (_or_value := getattr(base, "__orig_bases__", None)):
                return self._search_for_generic_type(_or_value)

        return generic_type



"""Houses Option"""

from typing import Any
from dataclasses import dataclass, fields

@dataclass
class Option:
    """Option Type"""

    name: str
    value: Any

    def __post_init__(self):
        for field in fields(self):
            value = getattr(self, field.name)
            if field.type == Any:
                return
            if not isinstance(value, field.type):
                raise ValueError(f'Expected {field.name!r} to be {field.type}, '
                                f'got {type(value)}')



"""Houses Abstract Request"""

from __future__ import annotations
from abc import abstractmethod, ABC
from typing import List, Union, TypeVar, Generic, Optional, Iterable, Dict, Any

from requests import Response

from pyrestsdk import AbstractServiceClient

from pyrestsdk.type.enum import HttpsMethod

from pyrestsdk.type.model import (
    CommonBase,
    QueryOption,
    HeaderOption,
    HeaderOptionCollection,
    QueryOptionCollection,
)

B = TypeVar("B", bound="AbstractRequest")
S = TypeVar("S", bound=AbstractServiceClient)
O = TypeVar("O", QueryOption, HeaderOption)
T = TypeVar("T")


class AbstractRequest(
    CommonBase,
    Generic[T],
    ABC,
):
    """Abstract Request Type"""

    @abstractmethod
    def __init__(
        self: B, request_url: str, client: S, options: Optional[Iterable[O]]
    ) -> None:
        """Instantiates new request"""
        super().__init__()

    @property
    @abstractmethod
    def header_options(self) -> HeaderOptionCollection:
        """Gets the headers"""

    @property
    @abstractmethod
    def request_method(self) -> HttpsMethod:
        """Gets/Sets the https method"""

    @property
    @abstractmethod
    def query_options(self) -> QueryOptionCollection:
        """Gets the query options"""

    @property
    @abstractmethod
    def client(self: B) -> S:
        """Gets the Client"""

    @property
    @abstractmethod
    def request_url(self: B) -> str:
        """Gets/Sets the request URL

        Returns:
            str: The request URL
        """

    @abstractmethod
    def _parse_options(self, options: Optional[Iterable[O]]) -> None:
        """Parses the provided options into either header or query options"""

    @abstractmethod
    def _initialize_url(self, request_url: str) -> str:
        """Parses the query parameters from URL"""

    @abstractmethod
    def send_request(self, value: Optional[T]) -> Optional[Union[List[T], T]]:
        """Makes the desired request and returns the desired return type"""

    @abstractmethod
    def parse_response(
        self, _response: Optional[Response]
    ) -> Optional[Union[List[T], T]]:
        """Parses the response into the expected return"""

    @abstractmethod
    def _send_request(
        self, args: Dict[str, Any], value: Optional[Union[T, Dict[str, Any], str]]
    ) -> Optional[Response]:
        """Makes the desired request and returns Response or None"""

    @abstractmethod
    def _get_request_args(
        self, value: Optional[Union[T, Dict[str, Any], str]] = None
    ) -> Dict[str, Any]:
        """gets request arguments"""

    @abstractmethod
    def append_segment_to_request_url(self, url_segment: str) -> str:
        """Gets a URL that is the request builder's request URL with the segment appended.

        Args:
            url_segment (str): The segment to append to the request URL.

        Returns:
            str: A URL that is the request builder's request URL with the segment appended.
        """

    @abstractmethod
    def parse_exception(self, response: Response) -> None:
        """Raises exception based off response"""



"""Houses Header Option Collection"""
from typing import Optional, Dict

from pyrestsdk.type.model._options_collection import OptionsCollection
from pyrestsdk.type.model._header_option import HeaderOption


class HeaderOptionCollection(OptionsCollection[HeaderOption]):
    """Header Option Collection Type"""

    def __init__(self, headers: Optional[Dict[str, str]] = None) -> None:
        """Initializes a new Header Option Collection

        Args:
            headers (Optional[Dict[str, str]], optional):
            The headers to initialize the collection with. Defaults to None.
        """

        super().__init__()
        if headers is None:
            return None
        for key, value in headers.items():
            self.append(HeaderOption(key, value))
        return None

    def add(self, name: str, value: str) -> None: #pylint: disable=arguments-renamed
        """Adds a new header option

        Args:
            name (str): Header name
            value (str): Header value
        """
        self.append(HeaderOption(name=name, value=value))



"""Houses Header Option"""
from typing import Dict

from pyrestsdk.type.model._option import Option


class HeaderOption(Option):
    """Header Option Type"""

    @property
    def as_dict(self) -> Dict[str, str]:
        """Converts the Header Option to a dictionary

        Returns:
            Dict[str, str]: The dictionary representation
        """

        return {str(self.name): str(self.value)}

    def __str__(self) -> str:

        return f"{self.name}:{self.value}"



"""Houses Client Factory"""
from typing import TypeVar, List
from logging import getLogger

from requests import Session

from pyrestsdk.middleware import MiddlewarePipeline
from pyrestsdk.middleware import BaseMiddleware
from pyrestsdk.credential import AbstractBasicCredential
from pyrestsdk.clientfactory._abstract_client_factory import AbstractHTTPClientFactory

Logger = getLogger(__name__)

B = TypeVar("B", bound=BaseMiddleware)
C = TypeVar("C", bound=AbstractBasicCredential)


class HTTPClientFactory(AbstractHTTPClientFactory): #pylint: disable=too-few-public-methods
    """HTTP Client Factory"""

    def __init__(
        self,
        base_url: str,
        session: Session,
        protocol: str = "https",
    ) -> None:
        super().__init__(session)
        self._base_url = base_url
        self._protocol = protocol

    def create_with_custom_middleware(self, middleware: List[B]) -> Session:

        if not middleware:
            raise ValueError("Please provide a list of custom middleware")

        return self._register(middleware)

    def _set_base_url(self, url: str) -> None:

        Logger.info("%s._set_base_url(): method called", type(self).__name__)

        self.session.base_url = f"{self._protocol}://{url}.{self._base_url}" # type: ignore

        Logger.debug(
            "%s._set_base_url() : base url set to: %s",
            type(self),
            self.session.base_url, # type: ignore
        )

    def _register(self, middleware: List[B]) -> Session:
        """
        Helper method that constructs a middleware_pipeline with the specified middleware
        """

        Logger.info("%s._register: method called", type(self))

        if middleware:
            middleware_pipeline = MiddlewarePipeline()
            for ware in middleware:
                middleware_pipeline.add_middleware(ware)

            self.session.mount(f"{self._protocol}://", middleware_pipeline)

        return self.session



"""
Service Client
==============

"""

from abc import abstractmethod, ABC

from requests import Session, Response

from pyrestsdk._abstract_service_client import AbstractServiceClient

class ServiceClient(AbstractServiceClient, ABC):
    """
    Service Client Type
    """

    _session: Session = None

    @property
    def session(self) -> Session:
        """Gets the current session

        Returns:
            Session: The current session
        """

        return self._session

    def get(self, url: str, **kwargs) -> Response:
        r"""Sends a GET request. Returns :class:`Response` object.
        :param url: URL for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.session.get(self._instance_url(url), **kwargs)

    def options(self, url, **kwargs) -> Response:
        r"""Sends a OPTIONS request. Returns :class:`Response` object.
        :param url: URL for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.session.options(self._instance_url(url), **kwargs)

    def head(self, url, **kwargs) -> Response:
        r"""Sends a HEAD request. Returns :class:`Response` object.
        :param url: URL for the new :class:`Request` object.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.session.head(self._instance_url(url), **kwargs)

    def post(self, url, data=None, json=None, **kwargs) -> Response:
        r"""Sends a POST request. Returns :class:`Response` object.
        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param json: (optional) json to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        return self.session.post(self._instance_url(url), data=data, json=json, **kwargs)

    def put(self, url, data=None, **kwargs) -> Response:
        r"""Sends a PUT request. Returns :class:`Response` object.
        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """

        return self.session.put(self._instance_url(url), data=data, **kwargs)

    def patch(self, url, data=None, **kwargs) -> Response:
        r"""Sends a PATCH request. Returns :class:`Response` object.
        :param url: URL for the new :class:`Request` object.
        :param data: (optional) Dictionary, list of tuples, bytes, or file-like
            object to send in the body of the :class:`Request`.
        :param \*\*kwargs: Optional arguments that ``request`` takes.
        :rtype: requests.Response
        """
        return self.session.patch(self._instance_url(url), data=data, **kwargs)

    def delete(self, url: str, **kwargs) -> Response:
        r"""Sends a DELETE request.
        
        Args:
            url (str): URL for the new :class:`Request` object.
            \*\*kwargs: Optional arguments that ``request`` takes.
        
        Returns:
            requests.Response: HTTP Response from request
        """
        return self.session.delete(self._instance_url(url), **kwargs)

    def _instance_url(self, url: str) -> str:
        """Joins session base URL with relative segement provided
        
        Args:
            url (str): The relative url from the base URL `/*/*` or `*/*`
        
        Returns:
            str: The Instance URL
        """

        return self.session.base_url + url if (url[0] == '/') else f"/{url}"  # type: ignore

    @staticmethod
    @abstractmethod
    def _get_session(*args, **kwargs) -> Session:
        """Gets the Session

        Returns:
            Session: The Session
        """



"""
Supports Base Invoke Request
============================
"""

from __future__ import annotations

from typing import Optional, Union, List, TypeVar

from abc import abstractmethod

from pyrestsdk.type.model import Entity
from pyrestsdk.request.supports_types._supports_types import SupportTypes
from pyrestsdk.request.supports_types._supports_generic_type import SupportsGenericType
from pyrestsdk.type.enum import HttpsMethod

T = TypeVar("T", bound=Entity)


class SupportsBaseInvokeRequest(SupportsGenericType[T], SupportTypes):
    """Supports Base Invoke Request Type"""

    @property
    @abstractmethod
    def input_object(self) -> Optional[T]:
        """Gets the input object for the request

        Returns:
            Optional[T]: _description_
        """
        raise NotImplementedError("input_object is not implemented")

    @property
    @abstractmethod
    def request_method(self) -> HttpsMethod:
        """Gets the Request HTTPS Method

        Returns:
            HttpsMethod: The Request HTTPS Method
        """
        raise NotImplementedError("request_method is not implemented")

    @abstractmethod
    def send(self, input_object: Optional[T] = None) -> Optional[Union[List[T], T]]:
        """Sends the request

        Args:
            input_object (Optional[T]): The input object to send. defaults to None

        Returns:
            Optional[Union[List[T], T]]: The response to the request
        """
        raise NotImplementedError("send is not implemented")

    @property
    @abstractmethod
    def invoke_request(self) -> Optional[Union[List[T], T]]:
        """Invokes the prepared request

        Returns:
            Optional[Union[List[T], T]]: The response to the request
        """

        raise NotImplementedError("invoke_request is not implemented")



"""Houses Supports Generic Type
"""

from typing import Type, TypeVar, Generic

from pyrestsdk.request.supports_types._supports_types import SupportTypes
from pyrestsdk.type.model._get_generic_type_mixin import GetGenericTypeMixin

T = TypeVar("T")
B = TypeVar("B", bound="SupportsGenericType")

class SupportsGenericType(SupportTypes, GetGenericTypeMixin, Generic[T]): #pylint: disable=too-few-public-methods
    """Supports Generic Type
    """

    _generic_type: Type[T]

    def __init__(self: B, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._generic_type = self._get_generic_type()

    @property
    def generic_type(self: B) -> Type[T]:
        """Gets the generic type"""

        return self._generic_type



"""Houses Query Option Collection"""

from typing import Any

from urllib.parse import urlencode

from pyrestsdk.type.model._options_collection import OptionsCollection
from pyrestsdk.type.model._query_option import QueryOption

class QueryOptionCollection(OptionsCollection[QueryOption]):
    """Query Option Collection Type"""

    def __str__(self) -> str:

        return urlencode(self.as_dict)

    def add(self, key: str, value: Any) -> None:
        """Adds Query Option to collection

        Args:
            key (str): Name of query option
            value (Any): Value of query option
        """
        self.append(QueryOption(key, value))



"""Houses Supports Invoke Collection Request
"""

from typing import TypeVar, List

from abc import ABC

from pyrestsdk.type.model import Entity
from pyrestsdk.type.exception import UnexpectedReturnType
from pyrestsdk.request.supports_types._supports_base_invoke_request import SupportsBaseInvokeRequest

from pyrestsdk.type.enum import HttpsMethod

T = TypeVar("T", bound=Entity)
S = TypeVar("S", bound="SupportsInvokeCollectionRequest")

class SupportsInvokeCollectionRequest(SupportsBaseInvokeRequest[T], ABC):
    """Supports Invoke Request
    
    Request supports invokation at later time
    """

    @property
    def invoke_request(self:S) -> List[T]:
        """Invokes Request Expecting list of value return type

        Raises:
            UnexpectedReturnType: Error when return type is other than generic list

        Returns:
            List[T]: The list of value
        """

        _return = self.send(self.input_object)

        if self.request_method in [HttpsMethod.POST, HttpsMethod.PUT]:
            return _return

        if not isinstance(_return, list) or _return is None:
            raise UnexpectedReturnType(type(_return), List[self.generic_type])

        return _return



"""Houses Token Authorization Handler"""

from pyrestsdk.middleware.authorizationhandler._authorization_handler import AuthorizationHandler


class TokenAuthorizationHandler(AuthorizationHandler):
    """Token Authorization Handler Type"""

    def _get_auth_header(self) -> str:
        return f"Bearer {self.credential.get_token().token}"



"""
Common Base
===========
The shared parent for all types, it's used to keep the
user from mutating the class outside of being inherited.
"""

from typing import Any, Callable
from functools import wraps
import inspect


def frozen(cls: Callable) -> Callable:
    """
    Decorator for freezing ability to set attributes outside of specific instances.
    """
    # Replace the class's __setattr__ method with the frozen version
    original_setattr = cls.__setattr__

    @wraps(cls.__setattr__)
    def setattr_frozen(self: Any, name: str, value: Any) -> None:
        caller = inspect.stack()[1]
        co_name = caller.function

        # Allow assignments at the module level and during __init__
        if co_name in ["<module>", "__init__"]:
            original_setattr(self, name, value)
            return

        # Prevent adding new attributes
        if not hasattr(self, name):
            raise AttributeError(
                f"You cannot add attributes to {self.__class__.__name__}"
            )

        if isinstance(getattr(self.__class__, co_name, None), property):
            original_setattr(self, name, value)
            return

        # Prevent setting value of protected attributes, except for property setters
        if name.startswith("_"):
            raise AttributeError(
                f"You cannot set value of protected attribute {self.__class__.__name__}.{name}"
            )

        # Calls the original __setattr__ function
        original_setattr(self, name, value)

    # Replace the class's __setattr__ with the frozen version
    setattr(cls, "__setattr__", setattr_frozen)
    return cls


@frozen
class CommonBase:  # pylint: disable=too-few-public-methods
    """Common Base Type"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)



"""Houses Kerbose Authorization Handler"""

from pyrestsdk.middleware.authorizationhandler._authorization_handler import (
    AuthorizationHandler,
)


class KerboseAuthorizationHandler(AuthorizationHandler):
    """Kerbose Authorization Handler Type"""

    def _get_auth_header(self) -> str:
        from requests_kerberos import ( # pylint: disable=import-outside-toplevel
            HTTPKerberosAuth,
            OPTIONAL,
        )

        kerbose_auth = HTTPKerberosAuth(
            mutual_authentication=OPTIONAL, principal=self.credential.get_principle()
        )

        return f"Kerberos {kerbose_auth}"



"""Houses Query Option Type"""

from pyrestsdk.type.model._option import Option


class QueryOption(Option): #pylint: disable=too-few-public-methods
    """Query Option Type"""

    def __str__(self) -> str:

        return f"{self.name}={self.value}"



"""Houses Basic Authorization Handler"""

from pyrestsdk.middleware.authorizationhandler._authorization_handler import AuthorizationHandler


class BasicAuthorizationHandler(AuthorizationHandler):
    """Basic Authoziation Handler Type"""

    def _get_auth_header(self) -> str:
        return f"Basic {self.credential.get_basic()}"



"""Py REST SDK
"""

from pyrestsdk._abstract_service_client import AbstractServiceClient
from pyrestsdk._service_client import ServiceClient
from pyrestsdk.type.enum._feature_usage_flag import FeatureUsageFlag

__all__ = ["AbstractServiceClient", "FeatureUsageFlag", "ServiceClient"]

__module_name__ = "pyrestsdk"



"""Houses Type Collection"""

from typing import MutableSequence, List, TypeVar, Iterator, Iterable

T = TypeVar("T")


class TypeCollection(MutableSequence[T]):
    """A collection type for storing elements with specific type constraints."""

    def __init__(self) -> None:
        """Initialize an empty TypeCollection.
        """

        self._internal_list: List[T] = []

    def __getitem__(self, index: int) -> T:
        """Get an item at the given index.

        Args:
            index (int): The index of the item

        Returns:
            T: The item
        """

        return self._internal_list[index]

    def __iter__(self) -> Iterator[T]:
        """Return an iterator over the elements in the collection.

        Yields:
            Iterator[T]: The iterator over the elements in the collection.
        """

        return iter(self._internal_list)

    def __len__(self) -> int:
        """Return the number of elements in the collection.

        Returns:
            int: The number of elements in the collection.
        """

        return len(self._internal_list)

    def __delitem__(self, index: int) -> None:
        """Delete an item at the given index.

        Args:
            index (int): The index
        """

        del self._internal_list[index]

    def __setitem__(self, index: int, value: T) -> None:
        """Set the value of an item at the given index.

        Args:
            index (int): The index
            value (T): The value
        """

        self._internal_list[index] = value

    def insert(self, index: int, value: T) -> None:
        """Insert an item at the given index.

        Args:
            index (int): The index
            value (T): The value
        """

        self._internal_list.insert(index, value)

    def __repr__(self) -> str:
        """Returns a string representation of the collection.

        Returns:
            str: A string representation of the collection.
        """

        return f"{type(self).__name__}({self._internal_list})"

    def append(self, value: T) -> None:
        """Append an item to the end of the collection.

        Args:
            value (T): Item to append
        """

        self._internal_list.append(value)

    def remove(self, value: T) -> None:
        """Remove the first occurrence of the given value from the collection.

        Args:
            value (T): The value to remove
        
        Raises:
            ValueError: if the value is not present.
        """

        self._internal_list.remove(value)

    def clear(self) -> None:
        """Clear all elements in the collection.
        """

        self._internal_list.clear()

    def count(self, value: T) -> int:
        """Returns the number of occurrences of the given value in the collection.

        Args:
            value (T): The value

        Returns:
            int: The number of occurrences of the given value in the collection.
        """

        return self._internal_list.count(value)

    def index(self, value: T, *args) -> int:
        """Return the first index of the given value in the collection.

        Args:
            value (T): The value
        
        Raises:
            ValueError: if the value is not present.

        Returns:
            int: The first index of the given value in the collection.
        """

        return self._internal_list.index(value, *args)

    def reverse(self) -> None:
        """Reverse the order of the elements in the collection.
        """

        self._internal_list.reverse()

    def sort(self, /, *args, **kwargs) -> None:
        """Sort the elements in the collection in ascending order.

        The sort is in-place (i.e. the collection itself is modified)
        and stable (i.e. the order of two equal elements is maintained).

        If a key function is given, apply it once to each list
        item and sort them, ascending or descending, according to their
        function values.

        The reverse flag can be set to sort in descending order.
        """
        self._internal_list.sort(*args, **kwargs)

    def extend(self, values: Iterable[T]) -> None:
        """Extend the collection by appending elements from the given iterable.

        Args:
            values (Iterable[T]): The iterable elements to append
        """

        self._internal_list.extend(values)



"""
Authorization Handler
=====================
The base for all Authorization Handlers
"""

from typing import Optional, TypeVar, Union, Tuple, Mapping
from requests import PreparedRequest, Response
from pyrestsdk.middleware import BaseAuthorizationHandler
from pyrestsdk.credential import AbstractCredential
from pyrestsdk.type.enum import FeatureUsageFlag

T = TypeVar("T", bound=AbstractCredential)
A = TypeVar("A", bound="AuthorizationHandler")

class AuthorizationHandler(BaseAuthorizationHandler):
    """Common Authorization Handler Type"""

    credential: T

    def __init__(self, credential: T, **kwargs) -> None:
        super().__init__(credential, **kwargs)

    def send(  # pylint: disable=too-many-arguments
        self: A,
        request: PreparedRequest,
        stream: bool = False,
        timeout: Optional[Union[float, Tuple[float, float], Tuple[float, None]]] = None,
        verify: bool = True,
        cert: Optional[
            Union[bytes, str, Tuple[Union[bytes, str], Union[bytes, str]]]
        ] = None,
        proxies: Optional[Mapping[str, str]] = None,
    ) -> Response:
        """Makes a network request if next is none, otherwise requests the next middleware to do

        Args:
            request (PreparedRequest): The network request

        Returns:
            Response: Response from network request
        """

        context = request.context  # type: ignore

        request.headers.update({"Authorization": self._get_auth_header()})

        context.set_feature_usage = FeatureUsageFlag.AUTH_HANDLER_ENABLED
        response = super().send(request, stream, timeout, verify, cert, proxies)

        # Token might have expired just before transmission, retry the request one more time
        if response.status_code == 401 and self.retry_count < 2:
            self.retry_count += 1
            return self.send(request, stream, timeout, verify, cert, proxies)
        return response

    def _get_auth_header(self: A) -> str:
        raise NotImplementedError("Subclasses must implement _get_auth_header method")



"""Requests"""
from sys import version_info
from pyrestsdk.request._request import Request

if version_info < (3,10):
    from pyrestsdk.request._base_request39 import BaseRequest
else:
    from pyrestsdk.request._base_request310 import BaseRequest #pylint: disable=[import-error,no-name-in-module,syntax-error]

__all__ = ["BaseRequest", "Request"]



"""Houses Base Request"""

from __future__ import annotations

from typing import TypeVar, Union, Optional, Iterable, Dict, Any
import logging

from requests import Response

from pyrestsdk.type.enum import HttpsMethod
from pyrestsdk.type.model import Entity, QueryOption, HeaderOption
from pyrestsdk.request._request import Request

Logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Entity)


class BaseRequest(Request[T]): # pylint: disable=too-many-ancestors
    """The Base Request Type"""

    def _parse_options(
        self, options: Optional[Iterable[Union[QueryOption, HeaderOption]]]
    ) -> None:
        """Parses the provided options into either header or query options"""

        Logger.info("%s._parse_options: function called", type(self).__name__)

        if options is None:
            return None

        for option in options:
            if isinstance(option, HeaderOption):
                self.header_options.append(option)
            elif isinstance(option, QueryOption):
                self.query_options.append(option)
            else:
                raise TypeError(
                    f"Unexpected type: {type(option)}, expected instance of HeaderOption or QueryOption" #pylint: disable=line-too-long
                )
        return None

    def _send_request(
        self, args: Dict[str, Any], value: Optional[Union[T, Dict[str, Any], str]]
    ) -> Optional[Response]:
        """Sends request

        Args:
            args (Dict[str, Any]): _description_
            value (Optional[Union[T, Dict[str, Any], str]]): _description_

        Raises:
            Exception: _description_

        Returns:
            Optional[Response]: Response to request
        """

        Logger.info(
            "%s._send_request: %s request made",
            type(self).__name__,
            self.request_method.name,
        )

        match self.request_method: #pylint: disable=syntax-error
            case HttpsMethod.GET: #pylint: disable=syntax-error
                return self._client.get(**args)
            case HttpsMethod.POST: #pylint: disable=syntax-error
                return self._client.post(**args)
            case HttpsMethod.DELETE: #pylint: disable=syntax-error
                self._client.delete(**args)
                return None
            case HttpsMethod.PUT: #pylint: disable=syntax-error
                return self._client.put(**args)
            case _: #pylint: disable=syntax-error
                raise TypeError(f"Unknown HTTPS method {self.request_method.name}")



"""Houses Request Type
"""

from typing import TypeVar, List, Union, Optional, Iterable, Dict, Any
import logging
from urllib.parse import urlparse
import json

from pyrestsdk import AbstractServiceClient
from pyrestsdk.type.enum import HttpsMethod
from pyrestsdk.type.model import Entity, QueryOption, HeaderOption
from pyrestsdk.request._abstract_request import AbstractRequest
from pyrestsdk.request.supports_types import (
    SupportsGenericType,
    SupportsQueryOptions,
    SupportsHeaderOptions,
)

T = TypeVar("T", bound=Entity)
S = TypeVar("S", bound=AbstractServiceClient)

Logger = logging.getLogger(__name__)


class Request( #pylint: disable=too-many-ancestors
    SupportsHeaderOptions,
    SupportsQueryOptions,
    SupportsGenericType,
    AbstractRequest[T],
):
    """Request Type"""

    def __init__(
        self,
        request_url: str,
        client: S,
        options: Optional[Iterable[Union[QueryOption, HeaderOption]]],
    ) -> None:
        super().__init__(request_url, client, options)

        self._request_url: str = request_url
        self._client: S = client
        self._method: HttpsMethod = HttpsMethod.GET
        self._request_url: str = self._initialize_url(request_url)
        self._parse_options(options)

    @property
    def request_method(self) -> HttpsMethod:
        """Gets/Sets the HTTPS method

        Returns:
            HttpsMethod: The HTTPS method
        """

        return self._method

    @request_method.setter
    def request_method(self, value: HttpsMethod) -> None:
        self._method = value
        Logger.info("%s.Method: request method set to %s", type(self).__name__, value)

    @property
    def request_url(self) -> str:
        """Gets/Sets the request URL

        Returns:
            str: The request URL
        """

        return self._request_url

    @request_url.setter
    def request_url(self, value: str) -> None:
        self._request_url = value
        Logger.info("%s.request_url: request URL set to %s", type(self).__name__, value)

    @property
    def client(self) -> S:
        """Gets the Client"""

        return self._client

    def _initialize_url(self, request_url: str) -> str:
        """Parses the query parameters from URL"""

        Logger.info("%s._initialize_url: function called", type(self).__name__)

        if not request_url:
            raise TypeError("request url can't be None")

        url = urlparse(request_url)

        if url.query:
            self._split_query_options(url.query)

        return url._replace(query="").geturl()

    def send(
        self, value: Optional[Union[T, Dict[str, Any]]] = None
    ) -> Optional[Union[List[T], T]]:
        """Makes the desired request and returns the desired return type"""

        Logger.info("%s.Send: method called", type(self).__name__)

        args = self._get_request_args(value)

        response = self._send_request(args, value)

        if response is None:
            return None

        try:
            Logger.debug("%s.Send: raising response for status", type(self).__name__)
            response.raise_for_status()
        except Exception: #pylint: disable=broad-exception-caught
            return self.parse_exception(response)
        return self.parse_response(response)

    def _parse_input_object(self, input_value: Union[T, Dict[str, Any], str]) -> str:
        """Converts input object into JSON

        Args:
            input_value (Union[T, Dict[str, Any], str]): The input object

        Returns:
            str: JSON version of object
        """

        if isinstance(input_value, str):
            return input_value

        if not isinstance(input_value, dict):
            input_value = input_value.as_dict

        return json.dumps(input_value)

    def _get_request_args( #pylint: disable=arguments-renamed
        self, input_value: Optional[Union[T, Dict[str, Any], str]] = None
    ) -> Dict[str, Any]:
        args = {
            "url": self.request_url,
            "headers": self.header_options.as_dict,
            "params": str(self.query_options),
        }

        if self.request_method in (HttpsMethod.POST, HttpsMethod.PUT):
            if input_value is None:
                raise ValueError(f"Missing data for {self.request_method.name} request.")

            args["data"] = self._parse_input_object(input_value)

        return args

    def append_segment_to_request_url(self, url_segment: str) -> str:
        """Gets a URL that is the request builder's request URL with the segment appended.

        Args:
            url_segment (str): The segment to append to the request URL.

        Returns:
            str: A URL that is the request builder's request URL with the segment appended.
        """

        if not url_segment.startswith("/"):
            url_segment = f"/{url_segment}"

        return f"{self.request_url}{url_segment}"



"""
MIME Types
==========
reference URL: https://www.iana.org/assignments/media-types/media-types.xhtml
"""

from enum import Enum, auto


class MIMECategoryEnum(str, Enum):
    """MIME Category Enum
    """

    def __new__(cls, value, *args, **kwargs):
        if not isinstance(value, (str, auto)):
            raise TypeError(
                f"Values of MIMECategoryEnum must be strings: {value!r} is a {type(value)}"
            )

        prefix = f"{cls.__name__.lower()}/"

        if isinstance(value, str) and not value.startswith(prefix):
            raise TypeError(
                f"Value of MIMECategoryEnum must begin with the designated prefix:{value!r} does not start with {prefix!r}" #pylint: disable=line-too-long
            )

        return super().__new__(cls, value, *args, **kwargs)

    def __str__(self) -> str:

        return self.value


class MimeType: #pylint: disable=too-few-public-methods
    """MIME Type"""

    class Text(MIMECategoryEnum):
        """Text MIME Type"""

        PLAIN = "text/plain"
        HTML = "text/html"
        CSS = "text/css"
        JAVASCRIPT = "text/javascript"
        CSV = "text/csv"

    class Application(MIMECategoryEnum):
        """Application MIME Type"""

        OCTET_STREAM = "application/octet-stream"
        PKCS12 = "application/x-pkcs12"
        X_WWW_FORM_URLENCODED = "application/x-www-form-urlencoded"
        JSON = "application/json"
        PDF = "application/pdf"
        XML = "application/xml"
        ZIP = "application/zip"
        GZIP = "application/gzip"

    class Image(MIMECategoryEnum):
        """Image MIME Type"""

        JPEG = "image/jpeg"
        PNG = "image/png"
        GIF = "image/gif"
        SVG_XML = "image/svg+xml"
        WEBP = "image/webp"

    class Audio(MIMECategoryEnum):
        """Audio MIME Type"""

        MIDI = "audio/midi"
        MPEG = "audio/mpeg"
        WEBM = "audio/webm"
        OGG = "audio/ogg"
        WAV = "audio/wav"

    class Video(MIMECategoryEnum):
        """Video MIME Type"""

        MPEG = "video/mpeg"
        MP4 = "video/mp4"
        OGG = "video/ogg"
        WEBM = "video/webm"
        QUICKTIME = "video/quicktime"



"""Houses Options Collection"""

from typing import (
    List,
    Dict,
    TypeVar,
    Any,
    Tuple,
    overload,
    Union,
    Iterable,
    Type,
    Optional,
)
from abc import abstractmethod

from pyrestsdk.type.model._type_collection import TypeCollection
from pyrestsdk.type.model._option import Option
from pyrestsdk.type.model._get_generic_type_mixin import GetGenericTypeMixin

O = TypeVar("O", bound=Option)
OC = TypeVar("OC", bound="OptionsCollection")


class OptionsCollection(GetGenericTypeMixin, TypeCollection[O]):
    """Option Collection Type"""

    element_type: Type[O]

    def __init__(
        self, options: Optional[Union[Iterable[O], Dict[str, Any]]] = None
    ) -> None:
        super().__init__()
        self.element_type = self._get_generic_type()

        if options is None:
            return

        if isinstance(options, dict):
            for key, value in options.items():
                self.add(key, value)
            return

        for option in options:
            self.append(option)

    @property
    def options(self) -> Tuple[O]:
        """Gets the options

        Returns:
            Tuple[O]: The options
        """

        return tuple(self._internal_list)

    def __getitem__(self, index: int) -> Option:
        """Gets object at index"""

        return self._internal_list[index]

    def __len__(self) -> int:
        """gets the length of the collection"""

        return len(self._internal_list)

    def __delitem__(self, index: int) -> None:
        """Deletes object at index"""

        del self._internal_list[index]

    def __setitem__(self, index: int, value: O) -> None:
        """Sets object at index"""

        self._internal_list[index] = value

    @overload
    def extend(self, new_options: Dict[str, Any]) -> None: #pylint: disable=arguments-renamed
        ...

    @overload
    def extend(self, new_options: Iterable[O]) -> None: #pylint: disable=arguments-renamed
        ...

    def extend(self, new_options: Union[Iterable[O], Dict[str, Any]]) -> None: #pylint: disable=arguments-renamed
        cleaned_new_options = []

        if isinstance(new_options, dict):
            for key, value in new_options.items():
                print(self.element_type)
                cleaned_new_options.append(self.element_type(key, value))
        else:
            cleaned_new_options = new_options

        return super().extend(cleaned_new_options)

    @property
    def as_dict(self) -> Dict[str, Any]:
        """Gets the object as it's dictionary representation

        Returns:
            Dict[str, Any]: Object's dictionary representation
        """

        print(self._internal_list)

        return {option.name: option.value for option in self._internal_list}

    def as_list(self) -> List[O]:
        """Gets the object as it's list representation"""

        return self._internal_list

    def remove(self, option_name: str) -> None: #pylint: disable=arguments-renamed
        """Remove the option with the given name from the collection.

        Args:
            option_name (str): The name of the option to remove.

        Raises:
            ValueError: If the option with the given name is not present in the collection.
        """

        for i, option in enumerate(self._internal_list):
            if option.name == option_name:
                del self[i]
                return

        raise ValueError(f"Option '{option_name}' is not present in the collection.")

    def get(self, option_name: str, default: Any = None) -> Any:
        """Get the value of the HeaderOption with the specified name.

        Args:
            option_name (str): The name of the HeaderOption to get the value of.
            default (Any, optional): The default value to return if the HeaderOption is not found.

        Returns:
            Any: The value of the HeaderOption,
            or the default value if the HeaderOption is not found.
        """

        for option in self._internal_list:
            if option.name == option_name:
                return option.value

        return default

    def set(self, name: str, value: Any) -> None:
        """Set the value of an option with a given name.

        Args:
            name (str): The name of the option
            value (Any): The value to set
        """
        for option in self._internal_list:
            if option.name == name:
                option.value = value
                return

        _value = self.element_type(name, value)

        # If the option was not found, add it as a new option
        self.append(self.element_type(name, value))

    @abstractmethod
    def add(self, key: str, value: Any) -> None:
        """Adds new option to the collection

        Args:
            key (str): The key for the option
            value (Any): The value for the option
        """
